// Parallel Port to Atmel In-System-Programming. 1/2001.
// Copyright (C) 2001 by Alan K Biocca  Alan.Biocca@usa.net
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details. (www.gnu.org)
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


// Designed to program Atmel AVR devices using ISP connected directly to
// a pc's parallel port, and be reconfigurable for different connection
// schemes. the program is layered to improve maintenance characteristics.
//
// Note that this initial version requires only single byte at a time
// programming as none of the initial devices have page mode programming.
// buffer set at 8K


// notes and future plans

// This program is written in 'pascal order', functions are defined before
// they are used. I don't usually write this way, but for a small program
// it avoids redefining all the functions. This was a small program,
// though perhaps it isn't quite anymore.

// what happens when we read a protected chip? do we get FFs or ??
// how to differentiate flash from eeprom on stdout?
// have format intel/motorola option change output from std hex to ...

// note that different models of the chips have different 'while prog'
// values for readback type programming.. however as long as it is 7F or FF
// the current algorithm will work. FF values are not programmed anyway,
// and the 7Fs will merely use the timer, so they still work. similar for
// eeprom, only 80 and 7F are both possible with 2313 chip.


// change log

// beta 0.6 1/17/2001
//	 reads signature and flash, and writes flash program

// beta 0.6 1/18/2001
//	 has new fast self-cal delay function

// beta 0.7
//
// added free format hex input. takes lines with address data ...
// address is hex, arbitrary size, terminated by whitespace
// data is bytes, either 1-2 chars each with whitespace
// or 2 bytes fixed with no whitespace
// examples
//   aaaa dd dd dd dd dd
//   aaa d dd d d dd d dd
//   aa dddddddddddddddd
//
// added lpr selection lpr0 1 2
// changed flag letters/numbers
// added eeprom readout and write
// changed default chipsize to 2k/128 flash/eeprom for reading, writing
// does whatever is in the buffer from the file
//
// chip now recognizes a few chip ids and sets the max flash and
// eeprom mem avail. 2313 8515 8535 tiny22 so far.
// sizes default to 8192 and 512 if size is not understood.
//
// writing flash and eeprom will only try to the limit of the
// known memory, or the limit of the request file, whichever
// is smaller. However since the default is 8192/512 if the
// chip is not recognized it could write up to the larger size,
// if the file has addresses that high. If the
// chip is actually smaller, and the file contains addresses beyond
// the physical chip, it will wrap around in the chip. This has different
// effects depending on the type of memory it is. Eeprom will erase and
// rewrite over the lower bytes. Flash will try to write, but it can
// only set to zero bits that were ones, or were already zeroes. It cannot
// raise existing zeroes to ones. So it will timeout while waiting for
// the location to finish writing.
//
// added a usage computation to the writing sections. this shows the percent
// of the resource based on the address of the last location written. note
// that in the case of eeprom the location will not be written if it is
// already the correct value, so writing twice will produce no changes,
// and no report of usage the second time.

// beta 0.8
//
// added kanda mode
// added arbitrary io configuration --defio ...

// beta 0.82 1/27/2001
//
// added start/stop configuration bytes
// for things such as kanda (generalized)
//
// --definit NN[[[M]AA]B]
//
// NN is 8 bits for port+0 initial port config (enable) (all hex)
// M  is 4 for port+2
// AA is finish byte for port+0
// B  is finish 4 bits for port+2


#include <stdio.h>
#include <stdlib.h>

void pl(char *s) { printf("%s\n",s); }  // print line
void myexit(char *s);

void usage()
{
pl("");
pl("Programmable Parallel Port ISP programmer for Atmel AVR");
pl("Version Beta 0.82 Jan 27 2001 Copyright (c) by Alan.Biocca@usa.net");
pl("default configured for Dontronics DT-006 - www.dontronics.com");
pl("");
pl("usage: wppisp [--typeopt [arg] ...] -[012srmcpPedCDWXYZS] [<File> ...]");
pl("");
pl(" options:");
pl("");
pl(" --dt006 configure for DT006 programmer (www.dontronics.com)(default)");
pl(" --kanda configure for Kanda STK200 dongle (www.kanda.com)");
pl(" --defio RiCiMiSi         define custom i/o wiring");
pl(" --definit SS[[[T]UU]V]   define custom port init");
pl(" 0 use LP0 port 0x3BC");
pl(" 1 use LP1 port 0x378 (default)");
pl(" 2 use LP2 port 0x278");
pl("");
pl(" s read chip signature to stdout");
pl(" r read chip flash prog mem to stdout");
pl(" m read chip eeprom data mem to stdout");
pl("");
pl(" c clear (erase prog & eeprom) chip");
pl(" p program hexfile to flash prog mem (pre-erase)");
pl(" P program hexfile to flash prog mem (non-erase)");
pl(" e program hexfile to eeprom (future)");
pl("");
//pl(" v verify hexfile against chip (after programming)(future)");
//pl("");
//pl(" i intel hex output format (input is autodetect)");
//pl(" m motorola srecord output format (future)");
pl("");
pl(" d display debug info");
pl(" C re-calibrate delay");
pl(" D delay test");
pl(" W toggle reset loop");
pl(" X toggle clock loop");
pl(" Y toggle mosi loop");
pl(" Z read miso loop");
//pl(" 5 toggle enable loop (kanda)(future)");
pl(" S slow way down (1ms/bit)");

myexit("");
}

// compiled with Borland C++ v5
// settings - standard dos 16 bit, no floating point
// tabstops 8, indent 4


// global defines

#define BUFSIZE (8*1024)    // buffer size for chip buffer


// global variables

int debug = 0;		// debug flag
int slow = 0;   	// go slow flag adds per-bit delays

short buf[BUFSIZE];      // buffer more than 8 bit to differentiate FF from -1
int f_size=8192;         // size of program flash memory for current chip
int e_size=512;          // size of eeprom data memory
int last_address;        // last address written
int sig[4] = {0,0,0,0};  // chip signature bytes (only using 3)


// timing

#include <time.h>

// the millisecond delay function is designed to exceed somewhat the
// delay to insure that chip timing is safe
// delays by calling clock system function repeatedly for a delay
//
// auto calibrating
// calibrates by watching transitions in the clock and computing the
// number of clock() calls per millisecond. this is about 25 for 
// a pentium 350.

unsigned mscal = 0, mscalmax = 0; int calmode = 0,tps;

void delayMs(unsigned ms)  // millisecond delay
{
    clock_t t,u,t1,t2; unsigned i,c,cm;

    if (ms > 1100) ms = 0;           // clamp period

    if (mscal == 0)                  // initialize delayMs
    {
        mscal = calmode = 10000;
        tps = CLK_TCK;

        for (i = 0; i < 7; ++i)
        {
            delayMs(1000);
        }
        calmode = 0;
        mscal = mscalmax + mscalmax/4;  // 25% insurance
        delayMs(ms);
        return;
    }

    t1 = t2 = 0;
    t = clock();

    while (ms--)
    {
        for (i = 0; i < mscal; ++i)
        {
            ++c;
            if (t != (u = clock()))
            {
                if (t1 == 0) t1 = u, c = 0;         // first edge
                else if (t2 == 0)
                {
                    if (calmode)
                    {
                        // found two edges, compute calibration
                        t = (u - t1) * 1000;
                        t /= tps;                // milliseconds between
                        cm = c / t;              // clocks per millisecond
                        if (debug)
                        {
                            printf("t %ld u %ld t1 %ld c %d cm %d max %d\n",
                                t,u,t1,c,cm,mscalmax);
                        }
                        if (mscalmax < cm) mscalmax = cm;

                        return;    // return early if cal mode
                    }
                }
                t = u;
            }
        }
    }
}

void delayCalibrate()  // initialize the delay calibration
{
    mscal = 0;
    delayMs(1);
}

void delayTest()  // test delay with 10 second count
{
    int i,tps;

    printf("10 second delay test\n");
    for (i = 0; i <= 10; ++i)
    {
        printf("%d\n",i);
        delayMs(1000);
    }
    tps = CLK_TCK;
    printf("clocks/ms %d ticks/ms %d\n", mscal, tps);
}


// hex conversion

int htoi(c)  // hex char to integer
char c;
{
    int r=-1;

    if (c >= '0' && c <= '9') r = c - '0';
    if (c >= 'a' && c <= 'f') r = c - 'a' + 10;
    if (c >= 'A' && c <= 'F') r = c - 'A' + 10;

    if (r >= 0) return r;

    myexit("nonhex char in hex field");  return 0;
}


// parallel port bit i/o

// requires dos.h, available in dos and win16 (win9x, winME)
// will not work on win32 (NT, 2000)
// adjust to use a parallal port driver for NT,2000 (future)

// borland c++ dos/win16 i/o functions (dos/win9x only)

#include <dos.h>

// defines i/o functions:
//
// unsigned char inportb(int port)
// void outportb(int port, unsigned char val);

#define LP1 0x378  // lp1 parallel printer port (default)
#define LP0 0x3bc  // lp0
#define LP2 0x278  // lp2

int ppbase = LP1;


// configurable bit i/o

// printer parallel port input-output

// parallel port documentation from various on the net
// mostly the Linux LDP Parallel IO Port Programming Mini-HOWTO

// base + 0 is data output, write to output pins
// bit  pin   name
// 0    2     LSB  (0=low, 1=high)
// 1    3
// 2    4
// 3    5
// 4    6
// 5    7
// 6    8
// 7    9     MSB

// output codes 0..7 ==> port base+0 data output bits 0..7

// base + 1 is status port, read-only
// bit  pin   name
// 3    15    Error In (1=high)
// 4    13    Select In (1=high)
// 5    12    Paper Error In (1=high)
// 6    10    ACK In (1=high)
// 7    11    Busy In (0=high)

// input codes 3..7 ==> port base+1 bit 3..7

// base + 2 is the control port, write-only
// bit  pin   name
// 0    1     STROBE (0=high)
// 1    14    AUTO_FD_XT (0=high)
// 2    16    INIT (1=high)
// 3    17    Select Out (0=high)
// 4          enable IRQ
// 5          extended mode direction (0 = write, 1 = read)

// output codes 8..B ==> port base+2 bits 0..3


// initial and final io configuration

int initb[4] = {0xFF,4,0xFF,4};  // initial and final bytes
int port0mem, port2mem;		 // port image memories
int init_io_flag = 0;  		 // io initialization status

void init_io()  // initialize for parallel io, set to initial state
{
    port0mem = initb[0];
    outportb(ppbase,port0mem);   

    port2mem = initb[1];
    outportb(ppbase+2,port2mem);  

    init_io_flag = 1;
}

void fin_io()  // terminate the io - put hw in finished state
{
    outportb(ppbase,initb[2]);   
    outportb(ppbase+2,initb[3]);  
}

void config_init(char *s)  // configure initialization from a hex string
{
    int i;

    if (debug) printf("config init %s\n", s);

    switch (strlen(s))
    {
        case 6:
    	    initb[3] = htoi(s[5]);
	case 5:
            initb[2] = htoi(s[3])*16+htoi(s[4]);
	case 3:
            initb[1] = htoi(s[2]);
	case 2:
            initb[0] = htoi(s[0])*16+htoi(s[1]);

	    break;

	default: myexit("config_init string length wrong");
    }

    if (debug)
    {
        printf("ini-fin bytes:");
        for (i = 0; i < 4; ++i) printf(" %02X", initb[i]);
        printf("\n");
    }
}


void outbit(int bitnum, int invert, int state)  // write a bit
{
    int andmask, ormask;

    if (init_io_flag == 0) init_io();

    state &= 1;				// use lsb

    ormask = 1 << (bitnum & 7);
    andmask = (~ormask) & 0xFF;

    if ((invert && state) || (!invert && !state)) ormask = 0;

    if (bitnum >= 0 && bitnum <= 7)
    {
	port0mem = (port0mem & andmask) | ormask;
	outportb(ppbase+0, port0mem);
    }
    else if (bitnum > 7 && bitnum <= 0xB)
    {
	port2mem = (port2mem & andmask & 0xF) | ormask;
	outportb(ppbase+2, port2mem);
    }
    else myexit("output bit address not in range");

    if (slow) delayMs(1);		// need a shorter delay perhaps 0.1 ms
}


// for some reason delays seem to be causing errors here
// would not expect this..

int inbit(int bitnum, int invert)  // read a bit
{
    int rbit;

    if (bitnum < 3 || bitnum > 7)
        myexit("input bit address not in range");

    rbit = inportb(ppbase+1) >> bitnum;

    if (invert) rbit = ~rbit;

    return rbit&1;
}


// configure i/o

// define port bits protocol
//
// use five character hex number
// <reset add><invert><clock adr><inv><mosi adr><inv><miso adr><inv>
//
// each pair is address for bit, then invert for that bit
//	zero = low  ==> noninverting ==> 0
//	zero = high ==> inverting    ==> 1
//
// DT-006 example:
//
//	reset is base+0 bit 2 noninverting which is 20
//	clock is base+0 bit 3 noninverting which is 30
//	mosi  is base+0 bit 1 noninverting which is 00
//	miso  is base+1 bit 7 inverting    which is 71
//
//  so 20301071 is the code for the DT-006


int range_fail(int in, int lo, int hi)  // check for out of range
{
    return in < lo || in > hi;
}

int reset_bit, clock_bit, mosi_bit, miso_bit;
int reset_invert, clock_invert, mosi_invert, miso_invert;

void show_config()
{
    printf("config reset %d %d, clock %d %d, mosi %d %d, miso %d %d\n",
        reset_bit, reset_invert, clock_bit, clock_invert,
        mosi_bit, mosi_invert, miso_bit, miso_invert);
}

void config_io(char *s)  // configure i/o
{
    int invert;

    if (strlen(s) != 8) myexit("config io must be 8 hex chars");

    // <reset adr><invert><clock adr><inv><mosi adr><inv><miso adr><inv>

    if (range_fail(reset_bit = htoi(s[0]), 0, 0xB))  // break into hex chars
        myexit("reset bit out of range");

    if (range_fail(reset_invert = htoi(s[1]), 0, 1))
        myexit("reset invert out of range");

    if (range_fail(clock_bit = htoi(s[2]), 0, 0xB))  // store in globals
        myexit("clock bit out of range");

    if (range_fail(clock_invert = htoi(s[3]), 0, 1))
        myexit("reset invert out of range");

    if (range_fail(mosi_bit = htoi(s[4]), 0, 0xB))
        myexit("mosi bit out of range");

    if (range_fail(mosi_invert = htoi(s[5]), 0, 1))
        myexit("reset invert out of range");

    if (range_fail(miso_bit = htoi(s[6]), 3, 7))
        myexit("miso bit out of range");

    if (range_fail(miso_invert = htoi(s[7]), 0, 1))
        myexit("reset invert out of range");

    if (reset_bit == clock_bit || 		     // final check
        reset_bit == mosi_bit  ||
	clock_bit == mosi_bit)    myexit("bits must differ");

    if (debug) show_config();
}

void set_dt006()  // setup for Dontronics DT-006 mode i/o
{
    if (debug) printf("Dontronics DT-006 mode\n");

    config_io("20300071");
}

void set_kanda()  // setup for kanda mode i/o
{
    // kanda type stk200 i/o dongle wiring
    //
    // pc pin bit function
    //
    // 4      2   0=enable B (Reset, MISO) (inverting)
    // 5      3   0=enable A (MOSI, CLOCK) (inverting)
    // 6      4   CLOCK to chip (noninverting)
    // 7      5   MOSI to chip (noninverting)
    // 9      7   RESET to chip (noninverting)
    // 10     6   MISO to host (noninverting)

    if (debug) printf("Kanda STK200 mode\n");

    // setup two enables

    //outbit(2, 1, 1);  // pin 4 lo
    //outbit(3, 1, 1);  // pin 5 lo

    // <reset><clock><mosi><miso> <invert>

    config_io("70405060");
    config_init("F3");  // pins 4,5 lo
}


// functional bit i/o

void set_sck_0()  // set clock low
{
    outbit(clock_bit, clock_invert, 0);
}

void set_sck_1()  // set clock high
{
    outbit(clock_bit, clock_invert, 1);
}

void assert_reset_0()  // assert reset low
{
    outbit(reset_bit, reset_invert, 0);
}

void release_reset()  // release reset (set high)
{
    outbit(reset_bit, reset_invert, 1);
}

void set_mosi(int b)  // set master out slave in bit to LSB
{
    outbit(mosi_bit, mosi_invert, b);
}

void release_io()  // release chip to run
{
    set_sck_1();
    set_mosi(1);
    release_reset();

    fin_io();				// release port
}

int read_miso()  // read master in slave out bit to LSB
{
    return inbit(miso_bit, miso_invert);
}



// test i/o

void toggleReset()  // toggle reset lo 1 second, hi 2 seconds, 5 times
{
    int i;

    printf("toggling RESET 5 cycles\n");
    for (i = 0; i < 5; ++i)
    {
        printf("L");
        assert_reset_0();
        sleep(1);
        printf("H");
        release_reset();
        sleep(2);
    }
}

void toggleClock()  // toggle clock lo 1 second, hi 2 seconds, 5 times
{
    int i;

    printf("toggling SCK 5 cycles\n");
    for (i = 0; i < 5; ++i)
    {
        printf("H");
	set_sck_1();
        sleep(1);
        printf("L");
	set_sck_0();
        sleep(2);
    }
}

void toggleMosi() // toggle mosi lo 1 second, hi 2 seconds, 5 times
{
    int i;

    printf("toggling MOSI 5 cycles\n");
    for (i = 0; i < 5; ++i)
    {
        printf("H");
	set_mosi(1);
        sleep(1);
        printf("L");
	set_mosi(0);
        sleep(2);
    }
}

void readMiso() // read miso each second 10 times
{
    int i;

    printf("reading MISO 10 times\n");
    for (i = 0; i < 10; ++i)
    {
	printf("miso: %d\n",read_miso());
        sleep(1);
    }
}


// functional compound i/o

#define SYNC_DELAY 40		// 20ms sync required

void reset_sync()  // synchronize with chip, assert reset low
{
    delayMs(1);
    set_sck_0();		// get into programming mode
    delayMs(1);
    release_reset();
    delayMs(1);                 // at least 2 clock cycle positive reset
    assert_reset_0();
    delayMs(SYNC_DELAY);        // min 20ms delay
}


// abort

void myexit(char *s)  // release chip and abort
{
    release_io();

    if (*s) printf("error abort: %s\n",s);
    exit(1);
}


// byte i/o

// send a byte and read one back simultaneously

// assumes clock low in, leaves clock low out
// assumes the sets have enough delay for normal i/o

int io_byte(int sbyte)  // send byte to chip, read byte same time
{
    int i, rbyte = 0, write_bit;

    for (i = 8; i ; --i)
    {
        write_bit = (sbyte & 128) >> 7;		// write bit
	set_mosi(write_bit);

	sbyte <<= 1;				// advance a bit

	set_sck_1();				// clock high

        rbyte <<= 1;
	rbyte |= read_miso();			// read bit & advance

	set_sck_0();				// clock low
    }

    return rbyte;
}

// write and read chip one 4 byte command sequence

int read1,read2,read3,read4;		// readback value globals

void io_4bytes(int b1, int b2, int b3, int b4)
{
    read1 = io_byte(b1);
    read2 = io_byte(b2);
    read3 = io_byte(b3);
    read4 = io_byte(b4);

    if (debug)
    {
        printf("io S %02X %02X %02X %02X, R %02X %02X %02X %02X\n",
            b1,b2,b3,b4,read1,read2,read3,read4);
    }
}

void send_program_enable()  // enable programming
{
    int i;

    reset_sync();
    
    for (i = 0; i < 32; ++i)
    {
        io_4bytes(0xAC,0x53,0,0);	// send program command

	if (read3 == 0x53) return;

        set_sck_1();   			// cycle clock
        delayMs(1);                     // insurance delay
        set_sck_0();
        delayMs(1);                     // insurance delay
    }

    printf("expected 0x53, received %0X\n", read3);

    myexit("program enable cycle failed 32 times (no chip?)");
}


// high level chip i/o

#define ERASE_DELAY 40		// 8-18 milliseconds specified

void erase_chip()  // erase flash and eeprom
{
    send_program_enable();

    io_4bytes(0xAC,0x80,0,0);   // erase command
    delayMs(ERASE_DELAY);     

    reset_sync();               // re-establish sync
    send_program_enable();	// redundant on purpose

    // check erasure?  (future)
}


// diagnostics

// chip signature recognition table

struct chipdata
{
    char *chip_name;             // text name
    unsigned char id[3];         // three byte id
    unsigned short flash_size;   // in K
    unsigned short eeprom_size;  // in bytes
}

chiptab[] =
{
    { "2313",   0x1E, 0x91, 0x01, 2, 128 },
    { "Tiny22", 0x1E, 0x91, 0x06, 2, 128 },
    { "8515",   0x1E, 0x93, 0x01, 8, 512 },
    { "8535",   0x1E, 0x93, 0x03, 8, 512 },
    { "4434",   0x1E, 0x92, 0x02, 4, 256 },
    { "Tiny11", 1,    0,    0,    1, 0   },
    { "Tiny12", 1,    0,    0,    1, 64  },
    { "Tiny15", 1,    0,    0,    1, 64  },
    { "Tiny28", 1,    0,    0,    2, 0   },
    { "1200",   1,    0,    0,    1, 64  },
    { "2323",   1,    0,    0,    2, 128 },
    { "2343",   1,    0,    0,    2, 128 },
    { "2333",   1,    0,    0,    2, 128 },
    { "4414",   1,    0,    0,    4, 256 },
    { "4433",   1,    0,    0,    4, 256 },
    { "8534",   1,    0,    0,    8, 512 },
    { "Mega83", 1,    0,    0,    8, 512 },
    { "",       0,    0,    0,    0, 0   }   // end marker
};

void read_chip_signature()  // read signature bytes to std out
{
    int i;

    send_program_enable();

    printf("chip signature: ");
    for (i = 0; i < 3; ++i)
    {
        io_4bytes(0x30,0,i,0);
        printf("%02X ",read4);
        sig[i] = read4;
    }
    printf("\n");
    for (i = 0; chiptab[i].flash_size; ++i)  // search for matching id
    {
        if (chiptab[i].id[0] == sig[0] &&
            chiptab[i].id[1] == sig[1] &&
            chiptab[i].id[2] == sig[2])
        {
            f_size = chiptab[i].flash_size*1024;
            e_size = chiptab[i].eeprom_size;
            printf("chip recognized: %s, flash: %d, eeprom: %d\n",
                chiptab[i].chip_name, f_size, e_size);
        }
    }
}

void id_chip()  // read signature if not already done
{
    if (sig[0] != 0x91) read_chip_signature();
}


// read/program eeprom memory

int read_eeprom_byte(int adr)
{
    // 1010 H000  0000 aaaa  bbbb bbbb  iiii iiii

    int b1,b2,b3,b4;

    b1 = 0xA0;	// lsb determines hi/lo byte

    b2 = (adr & 0xF00) >> 8;
    b3 = (adr & 0xFF);
    b4 = 0;
    io_4bytes(b1,b2,b3,b4);	// check return?

    return read4;
}

void read_eeprom_stdout()  // program memory to stdout
{
    int i,j,b,e;

    send_program_enable();
    id_chip();

    for (i = 0; i < e_size; i += 16)  // readout until a line of FFs found
    {
        printf("%04X ",i);
        for (e = j = 0; j < 16; ++j)  // 16 bytes per line
        {
            b = read_eeprom_byte(i+j);
            printf(" %02X",b);
            if (b == 0xFF) ++e;  // count empty locations
        }
        printf("\n");
        if (e == 16) break;      // quit on an empty line
    }
}

void write_eeprom_byte(int adr, int val)  // send a byte to program flash
{
    // 1100 H000  0000 aaaa  bbbb bbbb  iiii iiii

    int b1,b2,b3,b4;

    b1 = 0xC0;	// lsb determines hi/lo byte
    b2 = (adr & 0xF00) >> 8;
    b3 = (adr & 0xFF);
    b4 = val;
    io_4bytes(b1,b2,b3,b4);
}

#define PROG_DELAY 18			// 4-9 ms programming delay each write

void buffer_to_eeprom()  // write program buffer to flash
{
    int adr,val; clock_t t0;  long usage;

    send_program_enable();      	// set program mode
    id_chip();
    last_address = 0;

    for (adr = 0; adr < e_size; ++adr)
    {
        val = buf[adr];

	if (val >= 0 && val <= 255)	// prog bytes not default
	{
            if (val != read_eeprom_byte(adr))          // check for already
            {
	        write_eeprom_byte(adr,val);
                if (adr > last_address) last_address = adr;
    	        if (val == 0x7F) delayMs(PROG_DELAY);  // 7Fs need delay
                if (val == 0x80) delayMs(PROG_DELAY);  // 80s need delay

                t0 = clock();                          // verify write & timeout
                while (val != read_eeprom_byte(adr))
                {
                    if ((clock() - t0) > tps)          // 1 second timeout
                    {
                        printf("eeprom write fail at %X\n", adr);
                        myexit("");
                    }
                }
            }
	}
    }
    usage = last_address*100L;
    usage /= e_size;
    if (last_address) printf("last eeprom address written %d $%X = %ld%%\n",
                              last_address, last_address, usage);
}


// read/program 'program' flash memory

int read_progmem_byte(int adr)
{
    // 0010 H000  0000 aaaa  bbbb bbbb  iiii iiii

    int b1,b2,b3,b4;

    b1 = 0x20 + ((adr & 1)<<3);	// lsb determines hi/lo byte
    adr >>= 1;			// discard lsb
    b2 = (adr & 0xF00) >> 8;
    b3 = (adr & 0xFF);
    b4 = 0;
    io_4bytes(b1,b2,b3,b4);	// check return?

    return read4;
}

void read_progmem_stdout()  // program memory to stdout
{
    int i,j,b,e;

    send_program_enable();
    id_chip();

    for (i = 0; i < f_size; i += 16)  // readout until a line of FFs found
    {
        printf("%04X ",i);
        for (e = j = 0; j < 16; ++j)  // 16 bytes per line
        {
            b = read_progmem_byte(i+j);
            printf(" %02X",b);
            if (b == 0xFF) ++e;  // count empty locations
        }
        printf("\n");
        if (e == 16) break;      // quit on an empty line
    }
}

void write_progmem_byte(int adr, int val)  // send a byte to program flash
{
    // 0100 H000  0000 aaaa  bbbb bbbb  iiii iiii

    int b1,b2,b3,b4;

    b1 = 0x40 + ((adr & 1)<<3);	// lsb determines hi/lo byte
    adr >>= 1;			// discard lsb
    b2 = (adr & 0xF00) >> 8;
    b3 = (adr & 0xFF);
    b4 = val;
    io_4bytes(b1,b2,b3,b4);	// check return?
}

void buffer_to_flash()  // write program buffer to flash
{
    int adr,val; clock_t t0; long usage;

    send_program_enable();      	// set program mode
    id_chip();
    last_address = 0;

    for (adr = 0; adr < f_size; ++adr)
    {
        val = buf[adr];

	if (val >= 0 && val < 255)	// prog bytes less than FF
	{
	    write_progmem_byte(adr,val);
            if (adr > last_address) last_address = adr;
    	    if (val == 0x7F) delayMs(PROG_DELAY);  // 7Fs need delay

            t0 = clock();                          // verify write & timeout
            while (val != read_progmem_byte(adr))
            {
                if ((clock() - t0) > (tps/10))     // 0.1 second timeout
                {
                    printf("flash write fail at %X\n", adr);
                    myexit("");
                }
            }
	}
    }
    usage = last_address*100L;
    usage /= f_size;

    if (last_address) printf("last flash address written %d $%X = %ld%%\n",
                              last_address, last_address, usage);
}


// hex file i/o

#define BLKSIZE 16
#define LINSIZE 133

char line[LINSIZE];

// Intel Hex Summary
// index   desc
//
// 0         colon
// 1..2      number of data bytes, so msg is bytes*2+11 in length
// 3..6      addr
// 7..8      type
// 9..n-3    data
// n-2..n-1  checksum (sum of underlying bytes 1..n-3 + checksum == 0)
//
// n is num chars in line

int lines_decoded=0;

void decodeIhex(char *line)  // decode intel hex
{
    int i,len,addr,typ,n,byt,sum;

    for (n=0; n < LINSIZE; ++n)		// size line, cleanup newline
    {
	if (line[n] == '\n' || line[n] == '\r' || line[n] == 0)
	{
	    line[n] = 0;
	    break;
	}
    }

    for (sum=0, i=1; i < n-1; i+=2)	// check the checksum
    {
        sum = (sum + (htoi(line[i])<<4) + htoi(line[i+1]) )&255;
    }

    if (sum != 0) printf("bad checksum %d\n",sum), myexit("");

    ++lines_decoded;

    //len = (htoi(line[1])<<4) + htoi(line[2]);

    addr = (htoi(line[3])<<12) + (htoi(line[4])<<8) +
           (htoi(line[5])<<4)  +  htoi(line[6]);

    typ = (htoi(line[7])<<4) + htoi(line[8]);

    if (typ != 0) return;		// ignore non data types

    for (i=9; i < n-3; i+=2)		// load the data
    {
        byt = (htoi(line[i])<<4) + htoi(line[i+1]);

        if (addr < 0 || addr >= BUFSIZE) myexit("addr out of buffer");

        buf[addr++] = byt;
    }
}

// Motorola Hex Summary
//
// future

// character tests

int is_hex(char c)  // is arg a legal hex character
{
    if (c >= '0' && c <= '9') return 1;
    if (c >= 'a' && c <= 'f') return 1;
    if (c >= 'A' && c <= 'Z') return 1;
    return 0;
}

int is_blank(c)  // is arg a blank or tab
{
    return (c == ' ' || c == '\t');
}

void decodeFhex(char *line)  // read free format hex   AAAA DD DD DD DD DD ...
{
    int addr,byt; char *p; p = line;  addr = 0;

    while (is_hex(*p)) addr = addr*16 + htoi(*p++);    // read address

    while (is_blank(*p)) ++p;

    while (*p)
    {
        if (is_hex(*p)) byt = htoi(*p++);              // read 1-2 char hex
        if (is_hex(*p)) byt = byt*16 + htoi(*p++);

        if (addr < 0 || addr >= BUFSIZE)  myexit("addr out of buffer");

        buf[addr++] = byt;

        while (*p && !is_hex(*p)) ++p;                 // skip optional nonhex
    }
    ++lines_decoded;
}

void decodeSrec(char *line)  // decode line of motorola hex
{
    // future
    line[0] = 0;  // this to stop compiler warning
    myexit("Motorola S-Record Support not complete");
}

void readHexFile(char *fn)  // read hex file into buffer
{
    FILE *infile; int i;

    if ((infile = fopen(fn,"r")) == 0) myexit("couldn't open hexfile");

    for (i = 0; i < BUFSIZE; ++i) buf[i] = -1;	      // prep buffer

    while (fgets(line,LINSIZE,infile) != NULL)	      // process hex infile
    {
        if (line[0] == ':') decodeIhex(line);	      // intel hex type
	else if (line[0] == 'S') decodeSrec(line);    // motorola type
        else if (is_hex(line[0])) decodeFhex(line);   // free format hex
    }

    fclose(infile);

    printf("%d lines decoded\n", lines_decoded);
}


// will need this for intex hex output
//
//void wrline(FILE *fil, short *buf, int adr,len)  // write a line of intel hex
//{
//    int typ=0,j,sum;
//
//    fprintf(fil,":%02X%04X%02X",len,adr,typ);	// header
//    sum = len + ((adr>>8)&255) + (adr&255) + typ;
//
//    for (j=0; j < len; ++j)
//    {
//        fprintf(fil,"%02X",buf[adr+j]&255);
//	sum = (sum + buf[adr+j]) & 255;
//    }
//
//    fprintf(fil,"%02X\n",(-sum)&255);		// checksum, DOS line format
//}
//
//void writeHexFile()  // write hex to std out
//{
//    for (n=BUFSIZE-1; n > 0; --n)		// find end of data
//        if (buf[n] != -1) break;
//    ++n;					// n is one beyond eod
//
//    for (i=0; i <= n-BLKSIZE; i += BLKSIZE)	// write buffer to output hexfile a block at a time
//        wrline(outfile,buf,i,BLKSIZE);
//    if (i < n-1)
//        wrline(outfile,buf,i,n-i);
//}



// main coming soon

char *target=0, *hexfile=0, *arg;

int argeq(char *s)  // test argument against string
{
    return (strcmp(arg,s) == 0);
}

int program = 0;   // program chip flag

int main(int argc, char **argv)
{
    int argi,chi;	// argument and character indexes

    set_dt006();	// default to DT-006

    if (argc == 1)
    {
        printf("wppisp = Parallel Port In System Programmer for Atmel AVR\n");
        printf("1/2001 Alan.Biocca@usa.net\n");
        printf("win9x version\n");
        printf("use 'wppisp -h' for help\n");
        myexit("");
    }

    for (argi = 1; argi < argc; ++argi)  // decode args
    {
	arg = argv[argi];

        if (arg[0] == '-')
	{
	    for (chi = 1; arg[chi]; ++chi)
	    {
	        switch (arg[chi])
	        {
		    case 'h': usage(); break;
		    case 'C': delayCalibrate(); break;
                    case 'D': delayTest(); break;
                    case '0': ppbase = LP0; break;
                    case '1': ppbase = LP1; break;
                    case '2': ppbase = LP2; break;
		    case 'W': toggleReset(); break;
		    case 'X': toggleClock(); break;
		    case 'Y': toggleMosi(); break;
		    case 'Z': readMiso(); break;
                    case 's': read_chip_signature(); break;
                    case 'd': debug = !debug; break;
                    case 'S': slow = !slow; break;
                    case 'r': read_progmem_stdout(); break;
                    case 'm': read_eeprom_stdout(); break;
                    case 'c': erase_chip(); break;

                    case 'p': erase_chip();     // erase, then program
                    case 'P':			// non erase flash entry
                    case 'e':			// eeprom entry
                        if (++argi >= argc) myexit("insuff args");
                        readHexFile(argv[argi]);
                        if (arg[chi] == 'e') buffer_to_eeprom();
                        else buffer_to_flash();
                        break;

		    case '-':
		        if ((strcmp(arg,"--kanda")) == 0) set_kanda();
                        else if ((strcmp(arg,"--dt006")) == 0) set_dt006();
                        else if ((strcmp(arg,"--defio")) == 0)
                        {
                            if (++argi >= argc) myexit("missing io def");
                            config_io(argv[argi]);
                        }
                        else if ((strcmp(arg,"--definit")) == 0)
                        {
                            if (++argi >= argc) myexit("missing init def");
                            config_init(argv[argi]);
                        }
			else myexit("bad -- switch");
                        arg[chi+1] = 0;
		        break;

		    default: printf("unimplemented switch: %c\n", arg[chi]);
		             myexit("use 'wppisp -h' for help");
	        }
	    }
	}
	else myexit("extra args");
    }
    
    release_io();
    return 0;
}

// EOF
